#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.errors as errors

import octobot_trading.exchange_data.order_book.channel.order_book_updater as order_book_updater
import octobot_trading.util as util


class OrderBookUpdaterSimulator(order_book_updater.OrderBookUpdater):
    def __init__(self, channel, importer):
        super().__init__(channel)
        self.exchange_data_importer = importer
        self.exchange_name = self.channel.exchange_manager.exchange_name

        self.last_timestamp_pushed = 0
        self.time_consumer = None

    async def start(self):
        await self.resume()

    async def handle_timestamp(self, timestamp, **kwargs):
        try:
            for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
                order_book_data = (await self.exchange_data_importer.get_order_book_from_timestamps(
                    exchange_name=self.exchange_name,
                    symbol=pair,
                    inferior_timestamp=timestamp,
                    limit=1))[0]
                if order_book_data[0] > self.last_timestamp_pushed:
                    self.last_timestamp_pushed = order_book_data[0]
                    await self.push(pair, order_book_data[-1], order_book_data[-2])
        except errors.DatabaseNotFoundError as e:
            self.logger.warning(f"Not enough data : {e}")
            await self.pause()
            await self.stop()
        except IndexError as e:
            self.logger.warning(f"Failed to access order_book_data : {e}")

    async def pause(self):
        if self.time_consumer is not None:
            await util.get_time_channel(self).remove_consumer(self.time_consumer)

    async def stop(self):
        await util.stop_and_pause(self)

    async def resume(self):
        if self.time_consumer is None and not self.channel.is_paused:
            self.time_consumer = await util.get_time_channel(self).new_consumer(self.handle_timestamp)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

import pytest
from mock import patch
from datetime import datetime

from octobot_commons.asyncio_tools import wait_asyncio_next_cycle
from octobot_commons.constants import MSECONDS_TO_MINUTE
from octobot_commons.enums import TimeFrames, TimeFramesMinutes
from octobot_commons.tests.test_config import load_test_config
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.exchanges.exchanges import Exchanges
from octobot_trading.api.exchange import cancel_ccxt_throttle_task
from tests import event_loop

from tests.exchanges import cached_markets_exchange_manager

pytestmark = pytest.mark.asyncio


MS_TIMESTAMP = round((datetime.utcnow() - datetime(1970, 1, 1)).total_seconds() * 1000)


def get_constant_ms_timestamp():
    return MS_TIMESTAMP


class TestExchanges:
    @staticmethod
    async def init_default():
        return load_test_config()

    async def test_add_exchange(self):
        config = await self.init_default()

        async with cached_markets_exchange_manager(config, "binanceus"), \
                cached_markets_exchange_manager(config, "bybit", True):
            assert "binanceus" in Exchanges.instance().exchanges
            assert "bybit" in Exchanges.instance().exchanges
            assert "test" not in Exchanges.instance().exchanges

        cancel_ccxt_throttle_task()
        # let updaters gracefully shutdown
        await wait_asyncio_next_cycle()

    async def test_get_exchange(self):
        config = await self.init_default()

        async with cached_markets_exchange_manager(config, "binanceus") as exchange_manager_binance, \
                cached_markets_exchange_manager(config, "bybit") as exchange_manager_bybit:

            assert Exchanges.instance().get_exchanges_list("binanceus")[0].exchange_manager is exchange_manager_binance
            assert Exchanges.instance().get_exchanges_list("bybit")[0].exchange_manager is exchange_manager_bybit

            with pytest.raises(KeyError):
                assert Exchanges.instance().get_exchanges_list("test")

        cancel_ccxt_throttle_task()
        # let updaters gracefully shutdown
        await wait_asyncio_next_cycle()

    async def test_del_exchange(self):
        config = await self.init_default()

        async with cached_markets_exchange_manager(config, "binanceus") as exchange_manager_binance, \
                cached_markets_exchange_manager(config, "bybit") as exchange_manager_bybit:

            Exchanges.instance().del_exchange("binanceus", exchange_manager_binance.id)
            assert "binanceus" not in Exchanges.instance().exchanges
            Exchanges.instance().del_exchange("bybit", exchange_manager_bybit.id)
            assert "bybit" not in Exchanges.instance().exchanges

            Exchanges.instance().del_exchange("test", "")  # should not raise

            assert Exchanges.instance().exchanges == {}
        cancel_ccxt_throttle_task()
        # let updaters gracefully shutdown
        await wait_asyncio_next_cycle()

    async def test_get_all_exchanges(self):
        config = await self.init_default()

        async with cached_markets_exchange_manager(config, "binanceus") as exchange_manager_binance, \
                cached_markets_exchange_manager(config, "bybit") as exchange_manager_bybit:

            exchanges = Exchanges.instance().get_all_exchanges()
            assert exchanges[0].exchange_manager is exchange_manager_binance
            assert exchanges[1].exchange_manager is exchange_manager_bybit

        cancel_ccxt_throttle_task()
        # let updaters gracefully shutdown
        await wait_asyncio_next_cycle()

    async def test_ms_timestamp_operations(self):
        config = await self.init_default()
        async with cached_markets_exchange_manager(config, "bybit") as exchange_manager_bybit:

            if os.getenv('CYTHON_IGNORE'):
                await exchange_manager_bybit.stop()
                return

            exchange = exchange_manager_bybit.exchange
            with patch.object(exchange, 'get_exchange_current_time', new=get_constant_ms_timestamp):
                expected_ms_timestamp = MS_TIMESTAMP - TimeFramesMinutes[TimeFrames.ONE_HOUR] * MSECONDS_TO_MINUTE * 200
                assert exchange.get_candle_since_timestamp(TimeFrames.ONE_HOUR, count=200) == expected_ms_timestamp

                expected_ms_timestamp = MS_TIMESTAMP - TimeFramesMinutes[TimeFrames.ONE_WEEK] * MSECONDS_TO_MINUTE * 200
                assert exchange.get_candle_since_timestamp(TimeFrames.ONE_WEEK, count=200) == expected_ms_timestamp

                expected_ms_timestamp = MS_TIMESTAMP - TimeFramesMinutes[TimeFrames.ONE_MONTH] * MSECONDS_TO_MINUTE * 2000
                assert exchange.get_candle_since_timestamp(TimeFrames.ONE_MONTH, count=2000) == expected_ms_timestamp

                expected_ms_timestamp = MS_TIMESTAMP - TimeFramesMinutes[TimeFrames.ONE_MINUTE] * MSECONDS_TO_MINUTE * 10
                assert exchange.get_candle_since_timestamp(TimeFrames.ONE_MINUTE, count=10) == expected_ms_timestamp

                expected_ms_timestamp = MS_TIMESTAMP - TimeFramesMinutes[TimeFrames.ONE_MINUTE] * MSECONDS_TO_MINUTE * 0
                assert exchange.get_candle_since_timestamp(TimeFrames.ONE_MINUTE, count=0) == expected_ms_timestamp

                # 2000 months into the future
                expected_ms_timestamp = MS_TIMESTAMP + TimeFramesMinutes[TimeFrames.ONE_MONTH] * MSECONDS_TO_MINUTE * 2000
                assert exchange.get_candle_since_timestamp(TimeFrames.ONE_MONTH, count=-2000) == expected_ms_timestamp



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import mock
import pytest

import octobot_trading.personal_data as personal_data
import octobot_trading.personal_data.orders.groups.balanced_take_profit_and_stop_order_group as \
    balanced_take_profit_and_stop_order_group
import octobot_trading.personal_data.orders.groups.trailing_on_filled_tp_balanced_order_group as \
    trailing_on_filled_tp_balanced_order_group
import octobot_trading.personal_data.orders.order_util as order_util
from tests.exchanges import backtesting_config, backtesting_exchange_manager, fake_backtesting


@pytest.fixture
def toftpb_group(backtesting_exchange_manager):
    orders_manager = mock.Mock()
    orders_manager.get_order_from_group = mock.Mock()
    return personal_data.TrailingOnFilledTPBalancedOrderGroup("name",  orders_manager)


@pytest.fixture
def side_balance():
    return balanced_take_profit_and_stop_order_group.SideBalance(None, False)


@pytest.fixture
def trailing_side_balance():
    return trailing_on_filled_tp_balanced_order_group.TrailingSideBalance(None, True)


@pytest.fixture
def order():
    trader = mock.Mock(exchange_manager=mock.Mock())
    target_order = personal_data.Order(trader)
    return target_order


def test_balances_factory(toftpb_group):
    closed_orders = []
    filled = False
    balances = toftpb_group._balances_factory(closed_orders, filled)
    assert isinstance(balances[toftpb_group.TAKE_PROFIT], balanced_take_profit_and_stop_order_group.SideBalance)
    assert isinstance(balances[toftpb_group.STOP], trailing_on_filled_tp_balanced_order_group.TrailingSideBalance)


def test_TrailingSideBalance_get_order_update(trailing_side_balance, order):
    # no trailing profile
    assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
        personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
        personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
        personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: None,
    }

    # incompatible trailing profile
    order.trailing_profile = mock.Mock()
    assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
        personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
        personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
        personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: None,
    }

    order.trailing_profile = personal_data.FilledTakeProfitTrailingProfile([
        personal_data.TrailingPriceStep(price, price, True)
        for price in (10000, 12000, 13000)
    ])
    # filled price does not trigger trailing: trailing trigger price is not reached
    closed_order = personal_data.Order(order.trader)
    closed_order.origin_price = decimal.Decimal(9000)
    trailing_side_balance.closed_order = closed_order
    with mock.patch.object(
        order_util, "get_potentially_outdated_price", mock.Mock(return_value=(decimal.Decimal(8900), True))
    ) as get_potentially_outdated_price_mock:
        assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
            personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: None,
        }
        get_potentially_outdated_price_mock.assert_called_once()

    # filled price triggers trailing: trailing trigger price is reached by closed order even though current price is bellow
    closed_order = personal_data.Order(order.trader)
    closed_order.origin_price = decimal.Decimal(10000)
    trailing_side_balance.closed_order = closed_order
    with mock.patch.object(
        order_util, "get_potentially_outdated_price", mock.Mock(return_value=(decimal.Decimal(8900), True))
    ) as get_potentially_outdated_price_mock:
        assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
            personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: decimal.Decimal(10000),
        }
        get_potentially_outdated_price_mock.assert_called_once()

    # filled price does not trailing: trailing trigger price is reached by current price even though closed order is bellow
    # BUT trailing threshold has already been reached (trailed already)
    closed_order = personal_data.Order(order.trader)
    closed_order.origin_price = decimal.Decimal(1000)
    trailing_side_balance.closed_order = closed_order
    with mock.patch.object(
        order_util, "get_potentially_outdated_price", mock.Mock(return_value=(decimal.Decimal(10000), True))
    ) as get_potentially_outdated_price_mock:
        assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
            personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: None,
        }
        get_potentially_outdated_price_mock.assert_called_once()

    # filled price triggers trailing: trailing trigger price is reached by current price even though closed order is bellow
    closed_order = personal_data.Order(order.trader)
    closed_order.origin_price = decimal.Decimal(1000)
    trailing_side_balance.closed_order = closed_order
    with mock.patch.object(
        order_util, "get_potentially_outdated_price", mock.Mock(return_value=(decimal.Decimal(12000), True))
    ) as get_potentially_outdated_price_mock:
        assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
            personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: decimal.Decimal(12000),
        }
        get_potentially_outdated_price_mock.assert_called_once()
